# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from ipywidgets import Layout

default_layout = Layout(width='100%')
default_style = {'description_width': 'auto'}
